from __future__ import annotations

from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, NotRequired, TypedDict

from pytest import mark, param

from utilities.typing import (
    eval_typed_dict,
    get_args,
    is_dict_type,
    is_frozenset_type,
    is_list_type,
    is_literal_type,
    is_mapping_type,
    is_namedtuple_class,
    is_namedtuple_instance,
    is_optional_type,
    is_sequence_type,
    is_set_type,
    is_union_type,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class TestEvalTypedDict:
    def test_main(self) -> None:
        class Example(TypedDict):
            a: int
            b: str
            c: NotRequired[float]

        result = eval_typed_dict(Example)
        expected = {
            "a": int,
            "b": str,
            "c": NotRequired[float],  # pyright: ignore[reportInvalidTypeForm]
        }
        assert result == expected

    def test_nested(self) -> None:
        class Outer(TypedDict):
            a: int
            b: str
            inner: Inner

        class Inner(TypedDict):
            c: int
            d: str

        result = eval_typed_dict(Outer, globals_=globals(), locals_=locals())
        expected = {"a": int, "b": str, "inner": {"c": int, "d": str}}
        assert result == expected


class TestGetArgs:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(dict[int, int], (int, int)),
            param(frozenset[int], (int,)),
            param(int | None, (int,)),
            param(int | str, (int, str)),
            param(list[int], (int,)),
            param(Literal["a", "b", "c"], ("a", "b", "c")),
            param(Mapping[int, int], (int, int)),
            param(Sequence[int], (int,)),
            param(set[int], (int,)),
        ],
    )
    def test_main(self, *, obj: Any, expected: tuple[Any, ...]) -> None:
        result = get_args(obj)
        assert result == expected


class TestIsAnnotationOfType:
    @mark.parametrize(
        ("func", "obj", "expected"),
        [
            param(is_dict_type, dict[int, int], True),
            param(is_dict_type, list[int], False),
            param(is_frozenset_type, frozenset[int], True),
            param(is_frozenset_type, list[int], False),
            param(is_list_type, list[int], True),
            param(is_list_type, set[int], False),
            param(is_mapping_type, Mapping[int, int], True),
            param(is_mapping_type, list[int], False),
            param(is_literal_type, Literal["a", "b", "c"], True),
            param(is_literal_type, list[int], False),
            param(is_optional_type, int | None, True),
            param(is_optional_type, int | str, False),
            param(is_optional_type, list[int], False),
            param(is_sequence_type, Sequence[int], True),
            param(is_sequence_type, list[int], False),
            param(is_set_type, list[int], False),
            param(is_union_type, int | str, True),
            param(is_union_type, list[int], False),
        ],
    )
    def test_main(
        self, *, func: Callable[[Any], bool], obj: Any, expected: bool
    ) -> None:
        assert func(obj) is expected


class TestIsNamedTuple:
    def test_main(self) -> None:
        class Example(NamedTuple):
            x: int

        assert is_namedtuple_class(Example)
        assert is_namedtuple_instance(Example(x=0))

    def test_class(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        assert not is_namedtuple_class(Example)
        assert not is_namedtuple_instance(Example(x=0))



from __future__ import annotations

import datetime as dt
import re
from functools import partial
from re import escape
from typing import Any

from orjson import OPT_PASSTHROUGH_DATETIME, OPT_SORT_KEYS, dumps, loads

from utilities.whenever import (
    parse_date,
    parse_zoned_datetime,
    serialize_date,
    serialize_zoned_datetime,
)

_DATE_PREFIX = "[d]"
_DATETIME_PREFIX = "[t]"


def serialize2(obj: Any, /, *, fallback: bool = False) -> bytes:
    """Serialize an object."""
    return dumps(
        obj,
        default=partial(_serialize2_default, fallback=fallback),
        option=OPT_PASSTHROUGH_DATETIME | OPT_SORT_KEYS,
    )


def _serialize2_default(obj: Any, /, *, fallback: bool = False) -> str:
    if isinstance(obj, dt.datetime):
        ser = serialize_zoned_datetime(obj)
        return f"{_DATETIME_PREFIX}{ser}"
    if isinstance(obj, dt.date):
        ser = serialize_date(obj)
        return f"{_DATE_PREFIX}{ser}"
    if fallback:
        return str(obj)
    raise TypeError


def deserialize2(data: bytes, /) -> Any:
    """Deserialize an object."""
    return _object_hook(loads(data))


_DATE_PATTERN = re.compile(rf"^{escape(_DATE_PREFIX)}(.+)$")
_DATETIME_PATTERN = re.compile(rf"^{escape(_DATETIME_PREFIX)}(.+)$")


def _object_hook(obj: Any, /) -> Any:
    if isinstance(obj, str):
        if match := _DATE_PATTERN.search(obj):
            return parse_date(match.group(1))
        if match := _DATETIME_PATTERN.search(obj):
            return parse_zoned_datetime(match.group(1))
    if isinstance(obj, dict):
        return {k: _object_hook(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return list(map(_object_hook, obj))
    return obj


__all__ = ["deserialize2", "serialize2"]



from __future__ import annotations

from collections.abc import Callable
from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from functools import partial, wraps
from inspect import iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NoReturn,
    TypeVar,
    cast,
    overload,
)

import treelib
from treelib.exceptions import NodeIDAbsentError

from utilities.datetime import get_now
from utilities.functions import get_class_name
from utilities.sentinel import Sentinel, sentinel
from utilities.treelib import Tree, filter_tree
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Iterable
    from zoneinfo import ZoneInfo

    from utilities.treelib import Node
    from utilities.types import StrMapping

# types


_F = TypeVar("_F", bound=Callable[..., Any])
_T = TypeVar("_T")


@dataclass(kw_only=True, slots=True)
class _TracerData(Generic[_T]):
    trees: list[_TreeNodeData[_T]] = field(default_factory=list)
    tree: _TreeNodeData[_T] | None = None
    node: Node[NodeData[_T]] | None = None


@dataclass(kw_only=True, slots=True)
class NodeData(Generic[_T]):
    """A collection of data at each call."""

    module: str
    qualname: str
    args: tuple[Any, ...] | None = None
    kwargs: StrMapping | None = None
    start_time: dt.datetime
    end_time: dt.datetime | None = None
    outcome: Literal["success", "failure", "suppressed"] | None = None
    result: _T | Sentinel = sentinel
    error: Exception | None = None

    @property
    def desc(self) -> str:
        terms: list[Any] = []
        if (self.outcome == "failure") and (self.error is not None):
            terms.append(get_class_name(self.error))
        terms.append(self.duration)
        joined = ", ".join(map(str, terms))
        return f"{self.tag} ({joined})"

    @property
    def duration(self) -> dt.timedelta | None:
        return None if self.end_time is None else (self.end_time - self.start_time)

    @property
    def tag(self) -> str:
        return f"{self.module}:{self.qualname}"


_TreeNodeData = Tree[NodeData[_T]]


# context vars


_TRACER_CONTEXT: ContextVar[_TracerData[Any]] = ContextVar(
    "_CURRENT_TRACER_NODE", default=_TracerData()
)


@overload
def tracer(
    func: _F,
    /,
    *,
    time_zone: ZoneInfo | str = ...,
    add_args: bool = ...,
    pre_call: Callable[..., None] | None = ...,
    suppress: type[Exception] | tuple[type[Exception], ...] | None = ...,
    post_error: Callable[[NodeData[Any]], None] | None = ...,
    add_result: bool = ...,
    post_result: Callable[[NodeData[Any], Any], None] | None = ...,
) -> _F: ...
@overload
def tracer(
    func: None = None,
    /,
    *,
    time_zone: ZoneInfo | str = ...,
    add_args: bool = ...,
    pre_call: Callable[..., None] | None = ...,
    suppress: type[Exception] | tuple[type[Exception], ...] | None = ...,
    post_error: Callable[[NodeData[Any]], None] | None = ...,
    add_result: bool = ...,
    post_result: Callable[[NodeData[Any], Any], None] | None = ...,
) -> Callable[[_F], _F]: ...
def tracer(
    func: _F | None = None,
    /,
    *,
    time_zone: ZoneInfo | str = UTC,
    add_args: bool = False,
    pre_call: Callable[..., None] | None = None,
    suppress: type[Exception] | tuple[type[Exception], ...] | None = None,
    post_error: Callable[[NodeData[Any]], None] | None = None,
    add_result: bool = False,
    post_result: Callable[[NodeData[Any], Any], None] | None = None,
) -> _F | Callable[[_F], _F]:
    """Context manager for tracing function calls."""
    if func is None:
        result = partial(
            tracer,
            add_args=add_args,
            time_zone=time_zone,
            pre_call=pre_call,
            suppress=suppress,
            post_error=post_error,
            add_result=add_result,
            post_result=post_result,
        )
        return cast(Callable[[_F], _F], result)

    if iscoroutinefunction(func):

        @wraps(func)
        async def wrapped_async(*args: Any, **kwargs: Any) -> Any:
            node_data, tree, tracer_data, token = _initialize(
                func, args, kwargs, time_zone=time_zone, add_args=add_args
            )
            if pre_call is not None:
                pre_call(node_data, *args, **kwargs)
            try:
                result = await func(*args, **kwargs)
            except Exception as error:  # noqa: BLE001
                _handle_error(
                    node_data,
                    args,
                    kwargs,
                    error,
                    time_zone=time_zone,
                    suppress=suppress,
                    post_error=post_error,
                )
            else:
                return _handle_result(
                    node_data,
                    result,
                    time_zone=time_zone,
                    add_result=add_result,
                    post_result=post_result,
                )
            finally:
                _cleanup(tracer_data, token, tree=tree)

        return cast(Any, wrapped_async)

    @wraps(func)
    def wrapped_sync(*args: Any, **kwargs: Any) -> Any:
        node_data, tree, tracer_data, token = _initialize(
            func, args, kwargs, time_zone=time_zone, add_args=add_args
        )
        if pre_call is not None:
            pre_call(node_data, *args, **kwargs)
        try:
            result = func(*args, **kwargs)
        except Exception as error:  # noqa: BLE001
            _handle_error(
                node_data,
                args,
                kwargs,
                error,
                time_zone=time_zone,
                suppress=suppress,
                post_error=post_error,
            )
        else:
            return _handle_result(
                node_data,
                result,
                time_zone=time_zone,
                add_result=add_result,
                post_result=post_result,
            )
        finally:
            _cleanup(tracer_data, token, tree=tree)

    return cast(Any, wrapped_sync)


def get_tracer_trees() -> list[_TreeNodeData[Any]]:
    """Get the tracer trees."""
    return _TRACER_CONTEXT.get().trees


def set_tracer_trees(trees: Iterable[_TreeNodeData[Any]], /) -> None:
    """Set the tracer tree."""
    _ = _TRACER_CONTEXT.set(_TracerData(trees=list(trees)))


def _initialize(
    func: Callable[..., Any],
    args: tuple[Any, ...],
    kwargs: StrMapping,
    /,
    *,
    time_zone: ZoneInfo | str = UTC,
    add_args: bool = False,
) -> tuple[NodeData[_T], _TreeNodeData | None, _TracerData[_T], Token[_TracerData[_T]]]:
    node_data: NodeData[_T] = NodeData(
        module=func.__module__,
        qualname=func.__qualname__,
        start_time=get_now(time_zone=time_zone),
    )
    if add_args:
        node_data.args = args
        node_data.kwargs = kwargs
    tracer_data: _TracerData[_T] = _TRACER_CONTEXT.get()
    if (tree := tracer_data.tree) is None:
        tree_use = tracer_data.tree = Tree()
        tracer_data.trees.append(tree_use)
    else:
        tree_use = tree
    child = tree_use.create_node(
        tag=node_data.tag, parent=tracer_data.node, data=node_data
    )
    token = _TRACER_CONTEXT.set(
        _TracerData(trees=tracer_data.trees, tree=tree_use, node=child)
    )
    return node_data, tree, tracer_data, token


def _handle_error(
    node_data: NodeData[_T],
    args: tuple[Any, ...],
    kwargs: StrMapping,
    error: Exception,
    /,
    *,
    time_zone: ZoneInfo | str = UTC,
    suppress: type[Exception] | tuple[type[Exception], ...] | None = None,
    post_error: Callable[[NodeData[_T]], None] | None = None,
) -> NoReturn:
    node_data.args = args
    node_data.kwargs = kwargs
    if (suppress is not None) and isinstance(error, suppress):
        outcome = "suppressed"
    else:
        outcome = "failure"
    _set_end_time_and_outcome(node_data, outcome, time_zone=time_zone)
    node_data.error = error
    if post_error is not None:
        post_error(node_data)
    raise error


def _handle_result(
    node_data: NodeData[_T],
    result: _T,
    /,
    *,
    time_zone: ZoneInfo | str = UTC,
    add_result: bool = False,
    post_result: Callable[[NodeData[_T], _T], None] | None = None,
) -> _T:
    _set_end_time_and_outcome(node_data, "success", time_zone=time_zone)
    if add_result:
        node_data.result = result
    if post_result is not None:
        post_result(node_data, result)
    return result


def _set_end_time_and_outcome(
    node_data: NodeData[Any],
    outcome: Literal["success", "failure", "suppressed"],
    /,
    *,
    time_zone: ZoneInfo | str = UTC,
) -> None:
    node_data.end_time = get_now(time_zone=time_zone)
    node_data.outcome = outcome


def _cleanup(
    tracer_data: _TracerData,
    token: Token[_TracerData],
    /,
    *,
    tree: _TreeNodeData | None = None,
) -> None:
    if tree is None:
        tracer_data.tree = None
    _TRACER_CONTEXT.reset(token)


@overload
def filter_failures(tree: _TreeNodeData[_T], /) -> _TreeNodeData[_T] | None: ...
@overload
def filter_failures(
    tree: Iterable[_TreeNodeData[_T]], /
) -> list[_TreeNodeData[_T]]: ...
def filter_failures(
    tree: _TreeNodeData[_T] | Iterable[_TreeNodeData[_T]], /
) -> _TreeNodeData[_T] | list[_TreeNodeData[_T]] | None:
    """Filter a tree down to the failures."""
    if isinstance(tree, treelib.Tree):
        result = filter_tree(tree, data=lambda x: x.outcome == "failure")
        try:
            result[result.root]
        except NodeIDAbsentError:
            return None
        return result
    trees = list(map(filter_failures, tree))
    return [t for t in trees if t is not None]


__all__ = [
    "NodeData",
    "filter_failures",
    "get_tracer_trees",
    "set_tracer_trees",
    "tracer",
]



from __future__ import annotations

from asyncio import create_task, get_running_loop, sleep
from typing import TYPE_CHECKING

from hypothesis import HealthCheck, Phase, given, settings
from hypothesis.strategies import DataObject, booleans, data
from redis.asyncio import Redis

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from tests.test_orjson2 import _Object, objects
from utilities.functions import get_class_name
from utilities.hypothesis import int64s, text_ascii, yield_test_redis
from utilities.orjson import SerializeError, deserialize, serialize
from utilities.redis import (
    RedisHashMapKey,
    RedisKey,
    publish,
    subscribe,
    subscribe_messages,
    yield_redis,
)
from utilities.sentinel import SENTINEL_REPR, Sentinel, sentinel

if TYPE_CHECKING:
    from pytest import CaptureFixture

from pytest import raises


class TestPublishAndSubscribe:
    @given(
        data=data(),
        channel=text_ascii(min_size=1).map(
            lambda c: f"{get_class_name(TestPublishAndSubscribe)}_obj_ser_{c}"
        ),
        obj=objects,
    )
    @settings(
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_all_objects_with_serialize(
        self, *, capsys: CaptureFixture, data: DataObject, channel: str, obj: _Object
    ) -> None:
        async with yield_test_redis(data) as test:

            async def listener() -> None:
                async for msg in subscribe(
                    test.redis.pubsub(), channel, deserializer=deserialize
                ):
                    print(msg)  # noqa: T201

            task = create_task(listener())
            await sleep(0.05)
            _ = await publish(test.redis, channel, obj, serializer=serialize)
            await sleep(0.05)
            try:
                out = capsys.readouterr().out
                expected = f"{obj}\n"
                assert out == expected
            finally:
                _ = task.cancel()

    @given(
        data=data(),
        channel=text_ascii(min_size=1).map(
            lambda c: f"{get_class_name(TestPublishAndSubscribe)}_text_no_ser_{c}"
        ),
        text=text_ascii(min_size=1),
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_text_without_serialize(
        self, *, capsys: CaptureFixture, data: DataObject, channel: str, text: str
    ) -> None:
        async with yield_test_redis(data) as test:

            async def listener() -> None:
                async for msg in subscribe(test.redis.pubsub(), channel):
                    print(msg)  # noqa: T201

            task = create_task(listener())
            await sleep(0.05)
            _ = await publish(test.redis, channel, text)
            await sleep(0.05)
            try:
                out = capsys.readouterr().out
                expected = f"{text.encode()}\n"
                assert out == expected
            finally:
                _ = task.cancel()


class TestSubscribeMessages:
    @given(
        channel=text_ascii(min_size=1).map(
            lambda c: f"{get_class_name(TestSubscribeMessages)}_{c}"
        ),
        message=text_ascii(min_size=1),
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(
        self, *, capsys: CaptureFixture, channel: str, message: str
    ) -> None:
        client = Redis()

        async def listener() -> None:
            async for msg in subscribe_messages(client.pubsub(), channel):
                print(msg)  # noqa: T201

        task = get_running_loop().create_task(listener())
        await sleep(0.05)
        _ = await client.publish(channel, message)
        await sleep(0.05)
        try:
            out = capsys.readouterr().out
            expected = f"{{'type': 'message', 'pattern': None, 'channel': b'{channel}', 'data': b'{message}'}}\n"
            assert out == expected
        finally:
            _ = task.cancel()


class TestRedisHashMapKey:
    @given(data=data(), key=int64s(), value=booleans())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self, *, data: DataObject, key: int, value: bool) -> None:
        async with yield_test_redis(data) as test:
            hm_key = RedisHashMapKey(name=test.key, key=int, value=bool)
            assert await hm_key.hget(test.redis, key) is None
            _ = await hm_key.hset(test.redis, key, value)
            assert await hm_key.hget(test.redis, key) is value

    @given(data=data(), value=booleans())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_sentinel_key(self, *, data: DataObject, value: bool) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        async with yield_test_redis(data) as test:
            hm_key = RedisHashMapKey(
                name=test.key, key=Sentinel, value=bool, key_serializer=serializer
            )
            assert await hm_key.hget(test.redis, sentinel) is None
            _ = await hm_key.hset(test.redis, sentinel, value)
            assert await hm_key.hget(test.redis, sentinel) is value

    @given(data=data(), key=int64s())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_sentinel_value(self, *, data: DataObject, key: int) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        async with yield_test_redis(data) as test:
            hm_key = RedisHashMapKey(
                name=test.key,
                key=int,
                value=Sentinel,
                value_serializer=serializer,
                value_deserializer=deserializer,
            )
            assert await hm_key.hget(test.redis, key) is None
            _ = await hm_key.hset(test.redis, key, sentinel)
            assert await hm_key.hget(test.redis, key) is sentinel


class TestRedisKey:
    @given(data=data(), value=booleans())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_bool(self, *, data: DataObject, value: bool) -> None:
        async with yield_test_redis(data) as test:
            key = RedisKey(name=test.key, type=bool)
            assert await key.get(test.redis) is None
            _ = await key.set(test.redis, value)
            assert await key.get(test.redis) is value

    @given(data=data())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_sentinel_with_serialize(self, *, data: DataObject) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        async with yield_test_redis(data) as test:
            key = RedisKey(
                name=test.key,
                type=Sentinel,
                serializer=serializer,
                deserializer=deserializer,
            )
            assert await key.get(test.redis) is None
            _ = await key.set(test.redis, sentinel)
            assert await key.get(test.redis) is sentinel

    @given(data=data())
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_sentinel_without_serialize(self, *, data: DataObject) -> None:
        async with yield_test_redis(data) as test:
            key = RedisKey(name=test.key, type=Sentinel)
            with raises(
                SerializeError, match="Unable to serialize object of type 'Sentinel'"
            ):
                _ = await key.set(test.redis, sentinel)


class TestYieldClient:
    async def test_sync(self) -> None:
        async with yield_redis() as client:
            assert isinstance(client, Redis)



from __future__ import annotations

from pytest import raises

from utilities.errors import ImpossibleCaseError, RedirectErrorError, redirect_error


class TestImpossibleCaseError:
    def test_main(self) -> None:
        x = None
        with raises(ImpossibleCaseError, match=r"Case must be possible: x=None\."):
            raise ImpossibleCaseError(case=[f"{x=}"])


class TestRedirectError:
    def test_redirect(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        with raises(SecondError), redirect_error(FirstError, SecondError):
            raise FirstError

    def test_no_redirect(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        class ThirdError(Exception): ...

        with raises(FirstError, match=""), redirect_error(SecondError, ThirdError):
            raise FirstError

    def test_match_and_redirect(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        def run_test() -> None:
            new = SecondError("second")
            with redirect_error(FirstError, new, match="first"):
                msg = "first"
                raise FirstError(msg)

        with raises(SecondError, match="second"):
            run_test()

    def test_match_and_args_empty_error(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        def run_test() -> None:
            with redirect_error(FirstError, SecondError, match="match"):
                raise FirstError

        with raises(
            RedirectErrorError, match=r"Error must contain a unique argument; got .*\."
        ):
            run_test()

    def test_match_and_args_non_unique_error(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        def run_test() -> None:
            with redirect_error(FirstError, SecondError, match="match"):
                raise FirstError(1, 2)

        with raises(
            RedirectErrorError, match=r"Error must contain a unique argument; got .*\."
        ):
            run_test()

    def test_match_and_arg_not_string_error(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        def run_test() -> None:
            with redirect_error(FirstError, SecondError, match="match"):
                raise FirstError(None)

        with raises(
            RedirectErrorError, match=r"Error argument must be a string; got None\."
        ):
            run_test()

    def test_match_and_no_redirect(self) -> None:
        class FirstError(Exception): ...

        class SecondError(Exception): ...

        def run_test() -> None:
            with redirect_error(FirstError, SecondError, match="something else"):
                msg = "initial"
                raise FirstError(msg)

        with raises(FirstError, match="initial"):
            run_test()



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
    cast,
    overload,
)
from uuid import UUID, uuid4

from redis.asyncio import Redis
from redis.typing import EncodableT

from utilities.datetime import MILLISECOND, SECOND, duration_to_float, get_now
from utilities.errors import ImpossibleCaseError
from utilities.iterables import always_iterable
from utilities.types import Duration, ensure_int

if TYPE_CHECKING:
    import datetime as dt

    from redis.asyncio import ConnectionPool
    from redis.asyncio.client import PubSub
    from redis.typing import ResponseT

    from utilities.iterables import MaybeIterable


_K = TypeVar("_K")
_T = TypeVar("_T")
_V = TypeVar("_V")


class RedisMessageSubscribe(TypedDict):
    type: Literal["subscribe", "psubscribe", "message", "pmessage"]
    pattern: str | None
    channel: bytes
    data: bytes


class RedisMessageUnsubscribe(TypedDict):
    type: Literal["unsubscribe", "punsubscribe"]
    pattern: str | None
    channel: bytes
    data: int


_HOST = "localhost"
_PORT = 6379
_SUBSCRIBE_TIMEOUT = SECOND
_SUBSCRIBE_SLEEP = 10 * MILLISECOND


@dataclass(repr=False, kw_only=True, slots=True)
class TestRedis:
    """A container for a redis client; for testing purposes only."""

    redis: Redis
    timestamp: dt.datetime = field(default_factory=get_now)
    uuid: UUID = field(default_factory=uuid4)
    key: str


@dataclass(kw_only=True)
class RedisHashMapKey(Generic[_K, _V]):
    """A hashmap key in a redis store."""

    name: str
    key: type[_K]
    key_serializer: Callable[[_K], bytes] | None = None
    value: type[_V]
    value_serializer: Callable[[_V], bytes] | None = None
    value_deserializer: Callable[[bytes], _V] | None = None

    async def hget(self, redis: Redis, key: _K, /) -> _V | None:
        """Get a value from a hashmap in `redis`."""
        ser_key = self._serialize_key(key)  # skipif-ci-and-not-linux
        result = await cast(  # skipif-ci-and-not-linux
            Awaitable[Any], redis.hget(self.name, cast(Any, ser_key))
        )
        match result:  # skipif-ci-and-not-linux
            case None:
                return None
            case bytes() as data:
                if self.value_deserializer is None:
                    from utilities.orjson import deserialize

                    return deserialize(data)
                return self.value_deserializer(data)
            case _:  # pragma: no cover
                raise ImpossibleCaseError(case=[f"{redis=}"])

    async def hset(self, redis: Redis, key: _K, value: _V, /) -> int:
        """Set a value in a hashmap in `redis`."""
        ser_key = self._serialize_key(key)  # skipif-ci-and-not-linux
        if self.value_serializer is None:  # skipif-ci-and-not-linux
            from utilities.orjson import serialize

            ser_value = serialize(value)
        else:  # skipif-ci-and-not-linux
            ser_value = self.value_serializer(value)
        result = await cast(  # skipif-ci-and-not-linux
            Awaitable[int],
            redis.hset(self.name, key=cast(Any, ser_key), value=cast(Any, ser_value)),
        )
        return ensure_int(result)  # skipif-ci-and-not-linux

    def _serialize_key(self, key: _K, /) -> bytes:
        """Serialize the key."""
        if self.key_serializer is None:  # skipif-ci-and-not-linux
            from utilities.orjson import serialize

            return serialize(key)
        return self.key_serializer(key)  # skipif-ci-and-not-linux


@dataclass(kw_only=True)
class RedisKey(Generic[_T]):
    """A key in a redis store."""

    name: str
    type: type[_T]
    serializer: Callable[[_T], bytes] | None = None
    deserializer: Callable[[bytes], _T] | None = None

    async def get(self, redis: Redis, /) -> _T | None:
        """Get a value from `redis`."""
        match await redis.get(self.name):  # skipif-ci-and-not-linux
            case None:
                return None
            case bytes() as data:
                if self.deserializer is None:
                    from utilities.orjson import deserialize

                    return deserialize(data)
                return self.deserializer(data)
            case _:  # pragma: no cover
                raise ImpossibleCaseError(case=[f"{redis=}"])

    async def set(self, redis: Redis, value: _T, /) -> int:
        """Set a value in `redis`."""
        if self.serializer is None:  # skipif-ci-and-not-linux
            from utilities.orjson import serialize

            value_use = serialize(value)
        else:  # skipif-ci-and-not-linux
            value_use = self.serializer(value)
        return await redis.set(self.name, value_use)  # skipif-ci-and-not-linux


@overload
async def publish(
    redis: Redis, channel: str, data: _T, /, *, serializer: Callable[[_T], EncodableT]
) -> ResponseT: ...
@overload
async def publish(
    redis: Redis,
    channel: str,
    data: EncodableT,
    /,
    *,
    serializer: Callable[[EncodableT], EncodableT] | None = None,
) -> ResponseT: ...
async def publish(
    redis: Redis,
    channel: str,
    data: Any,
    /,
    *,
    serializer: Callable[[Any], EncodableT] | None = None,
) -> ResponseT:
    """Publish an object to a channel."""
    data_use = (  # skipif-ci-and-not-linux
        cast(EncodableT, data) if serializer is None else serializer(data)
    )
    return await redis.publish(channel, data_use)  # skipif-ci-and-not-linux


@overload
def subscribe(
    pubsub: PubSub,
    channels: MaybeIterable[str],
    /,
    *,
    deserializer: Callable[[bytes], _T],
    timeout: Duration | None = ...,
    sleep: Duration = ...,
) -> AsyncIterator[_T]: ...
@overload
def subscribe(
    pubsub: PubSub,
    channels: MaybeIterable[str],
    /,
    *,
    deserializer: None = None,
    timeout: Duration | None = ...,
    sleep: Duration = ...,
) -> AsyncIterator[bytes]: ...
async def subscribe(
    pubsub: PubSub,
    channels: MaybeIterable[str],
    /,
    *,
    deserializer: Callable[[bytes], _T] | None = None,
    timeout: Duration | None = _SUBSCRIBE_TIMEOUT,  # noqa: ASYNC109
    sleep: Duration = _SUBSCRIBE_SLEEP,
) -> AsyncIterator[Any]:
    """Subscribe to the data of a given channel(s)."""
    channels = list(always_iterable(channels))  # skipif-ci-and-not-linux
    messages = subscribe_messages(  # skipif-ci-and-not-linux
        pubsub, channels, timeout=timeout, sleep=sleep
    )
    if deserializer is None:  # skipif-ci-and-not-linux
        async for message in messages:
            yield message["data"]
    else:  # skipif-ci-and-not-linux
        async for message in messages:
            yield deserializer(message["data"])


async def subscribe_messages(
    pubsub: PubSub,
    channels: MaybeIterable[str],
    /,
    *,
    timeout: Duration | None = _SUBSCRIBE_TIMEOUT,  # noqa: ASYNC109
    sleep: Duration = _SUBSCRIBE_SLEEP,
) -> AsyncIterator[RedisMessageSubscribe]:
    """Subscribe to the messages of a given channel(s)."""
    channels = list(always_iterable(channels))  # skipif-ci-and-not-linux
    for channel in channels:  # skipif-ci-and-not-linux
        await pubsub.subscribe(channel)
    channels_bytes = [c.encode() for c in channels]  # skipif-ci-and-not-linux
    timeout_use = (  # skipif-ci-and-not-linux
        None if timeout is None else duration_to_float(timeout)
    )
    sleep_use = duration_to_float(sleep)  # skipif-ci-and-not-linux
    while True:  # skipif-ci-and-not-linux
        message = cast(
            RedisMessageSubscribe | RedisMessageUnsubscribe | None,
            await pubsub.get_message(timeout=timeout_use),
        )
        if (
            (message is not None)
            and (message["type"] in {"subscribe", "psubscribe", "message", "pmessage"})
            and (message["channel"] in channels_bytes)
            and isinstance(message["data"], bytes)
        ):
            yield cast(RedisMessageSubscribe, message)
        else:
            await asyncio.sleep(sleep_use)


@asynccontextmanager
async def yield_redis(
    *,
    host: str = _HOST,
    port: int = _PORT,
    db: str | int = 0,
    password: str | None = None,
    connection_pool: ConnectionPool | None = None,
    decode_responses: bool = False,
    **kwargs: Any,
) -> AsyncIterator[Redis]:
    """Yield an asynchronous redis client."""
    redis = Redis(
        host=host,
        port=port,
        db=db,
        password=password,
        connection_pool=connection_pool,
        decode_responses=decode_responses,
        **kwargs,
    )
    try:
        yield redis
    finally:
        await redis.aclose()


__all__ = [
    "RedisHashMapKey",
    "RedisKey",
    "TestRedis",
    "publish",
    "subscribe",
    "subscribe_messages",
    "yield_redis",
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from re import IGNORECASE, search
from subprocess import PIPE, CalledProcessError, check_output
from typing import TYPE_CHECKING, TypeVar, overload

from typing_extensions import override

from utilities.pathlib import PWD

if TYPE_CHECKING:
    from collections.abc import Callable

    from utilities.types import PathLike

_GET_BRANCH_NAME = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
_T = TypeVar("_T")
_U = TypeVar("_U")


def get_branch_name(*, cwd: PathLike = PWD) -> str:
    """Get the current branch name."""
    root = get_repo_root(cwd=cwd)
    output = check_output(_GET_BRANCH_NAME, stderr=PIPE, cwd=root, text=True)
    return output.strip("\n")


def get_repo_name(*, cwd: PathLike = PWD) -> str:
    """Get the repo name."""
    root = get_repo_root(cwd=cwd)
    output = check_output(
        ["git", "remote", "get-url", "origin"], stderr=PIPE, cwd=root, text=True
    )
    return Path(output.strip("\n")).stem  # not valid_path


def get_repo_root(*, cwd: PathLike = PWD) -> Path:
    """Get the repo root."""
    try:
        output = check_output(
            ["git", "rev-parse", "--show-toplevel"], stderr=PIPE, cwd=cwd, text=True
        )
    except CalledProcessError as error:
        # newer versions of git report "Not a git repository", whilst older
        # versions report "not a git repository"
        if search("fatal: not a git repository", error.stderr, flags=IGNORECASE):
            raise GetRepoRootError(cwd=cwd) from error
        raise  # pragma: no cover
    else:
        return Path(output.strip("\n"))


@dataclass(kw_only=True, slots=True)
class GetRepoRootError(Exception):
    cwd: PathLike

    @override
    def __str__(self) -> str:
        return f"Path is not part of a `git` repository: {self.cwd}"


@overload
def get_repo_root_or_cwd_sub_path(
    if_exists: Callable[[Path], _T], /, *, cwd: PathLike = ..., if_missing: None = ...
) -> _T | None: ...
@overload
def get_repo_root_or_cwd_sub_path(
    if_exists: Callable[[Path], _T],
    /,
    *,
    cwd: PathLike = ...,
    if_missing: Callable[[Path], _U] = ...,
) -> _T | _U: ...
def get_repo_root_or_cwd_sub_path(
    if_exists: Callable[[Path], _T],
    /,
    *,
    cwd: PathLike = PWD,
    if_missing: Callable[[Path], _U] | None = None,
) -> _T | _U | None:
    """Get a path under the repo root, if it exists, else under the CWD."""
    try:
        root = get_repo_root(cwd=cwd)
    except (FileNotFoundError, GetRepoRootError):
        if if_missing is None:
            return None
        return if_missing(Path(cwd))
    return if_exists(root)


__all__ = [
    "GetRepoRootError",
    "get_branch_name",
    "get_repo_name",
    "get_repo_root",
    "get_repo_root_or_cwd_sub_path",
]



from __future__ import annotations

from pytest import mark, param

from utilities.sys import VERSION_MAJOR_MINOR, _GetCallerOutput, get_caller


class TestGetCaller:
    @mark.parametrize(
        ("depth", "expected"),
        [param(1, "inner"), param(2, "outer"), param(3, "test_main")],
        ids=str,
    )
    def test_main(self, *, depth: int, expected: str) -> None:
        def outer() -> _GetCallerOutput:
            return inner()

        def inner() -> _GetCallerOutput:
            return get_caller(depth=depth)

        result = outer()
        assert result["module"] == "tests.test_sys"
        assert result["name"] == expected

    @mark.parametrize(
        ("depth", "expected"),
        [param(1, "inner"), param(2, "mid"), param(3, "outer"), param(4, "test_depth")],
        ids=str,
    )
    def test_depth(self, *, depth: int, expected: str) -> None:
        def outer() -> _GetCallerOutput:
            return mid()

        def mid() -> _GetCallerOutput:
            return inner()

        def inner() -> _GetCallerOutput:
            return get_caller(depth=depth)

        result = outer()
        assert result["module"] == "tests.test_sys"
        assert result["name"] == expected


class TestVersionMajorMinor:
    def test_main(self) -> None:
        assert isinstance(VERSION_MAJOR_MINOR, tuple)
        expected = 2
        assert len(VERSION_MAJOR_MINOR) == expected



from __future__ import annotations

from tenacity import retry, wait_fixed

from utilities.tenacity import before_sleep_log

# tenacity


_counter = 0


@retry(wait=wait_fixed(0.01), before_sleep=before_sleep_log())
def func_test_tenacity_before_sleep_log() -> int:
    global _counter  # noqa: PLW0603
    _counter += 1
    if _counter >= 3:
        return _counter
    raise ValueError(_counter)



from __future__ import annotations

from sys import _getframe, version_info
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from types import FrameType

VERSION_MAJOR_MINOR = (version_info.major, version_info.minor)


class _GetCallerOutput(TypedDict):
    module: str
    line_num: int
    name: str


def get_caller(*, depth: int = 2) -> _GetCallerOutput:
    """Get the calling function."""
    i = 0
    frame: FrameType | None = _getframe()  # pragma: no cover
    while (i < depth) and (frame.f_back is not None):
        i, frame = i + 1, frame.f_back
    return {
        "module": frame.f_globals["__name__"],
        "line_num": frame.f_lineno,
        "name": frame.f_code.co_name,
    }


__all__ = ["VERSION_MAJOR_MINOR", "get_caller"]
